@@ -1,5 +1,5 @@
 /*****************************************************************************
- * RRDtool 1.0.28  Copyright Tobias Oetiker, 1997 - 2000
+ * RRDtool 1.0.35  Copyright Tobias Oetiker, 1997 - 2000
  *****************************************************************************
  * rrd_restore.c  creates new rrd from data dumped by rrd_dump.c
  *****************************************************************************/
@@ -280,7 +280,7 @@ rrd_write(char *file_name, rrd_t *rrd)
       *rrd_file= *stdout;
     } else {
       if ((rrd_file = fopen(file_name,"wb")) == NULL ) {
-	rrd_set_error("can't create '%s'",file_name);
+	rrd_set_error("creating '%s': %s",file_name,strerror(errno));
 	rrd_free(rrd);
 	return(-1);
       }
